/*------------------------------------------------------------
IDEA 
  IDEA ou International Data Encryption Algorithm  um algoritmo 
  simtrico que utiliza chaves de 128 bits.
  Foi projetado em 1990 para ser de fcil programao, sendo 
  resistente a muitas formas de criptoanlise.
  
Funcionamento
  1) O algoritmo  usado tanto para a cifragem quanto para a 
     decifrao;
  2) Usa a confuso e a difuso misturando vrias operaes de 
     trs grupos algbricos diferentes: 
     -XOR (inverso de bits);
     -adio modular (inteiros de 16 bits, bigendian);
     -multiplicao modular (inteiros de 16 bits, bigendian).

Referencias:
  http://www.numaboa.com/content/view/336/120/
--------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "idea.h"

#define SIZEOF_UNSIGNED_LONG 4

// desative a linha abaixo se for usar em processadores
// bigendian (como motorola, Apple etc)
#define LITTLE_ENDIAN_HOST 1

typedef unsigned long  ulong;
typedef unsigned short ushort;
typedef unsigned char  byte;

typedef unsigned short u16;
typedef unsigned long  u32;

#define IDEA_KEYSIZE 16
#define IDEA_BLOCKSIZE 8
#define IDEA_ROUNDS 8
#define IDEA_KEYLEN (6*IDEA_ROUNDS+4)

typedef struct {
    u16 ek[IDEA_KEYLEN];
    u16 dk[IDEA_KEYLEN];
    int have_dk;
} IDEA_context;

void invert_key( u16 *ek, u16 dk[IDEA_KEYLEN] );
int cipher( byte *outbuf, byte *inbuf, u16 *key );
int do_setkey( IDEA_context *c, byte *key, int keylen );
void encrypt_block( IDEA_context *bc, byte *outbuf, byte *inbuf );
void decrypt_block( IDEA_context *bc, byte *outbuf, byte *inbuf );

//-------------------------------------------------------------------
// algoritmos matemticos envolvidos
//-------------------------------------------------------------------

u16 MUL(u16 x, u16 y) 
{    
     u32 p=x*y;    
     if (p == 0)       
        x = 65537-x-y;    
     else {       
          x = p >> 16;
          y = p;       
          x = y-x;       
          if (y < x) 
             x += 65537;    
     }    
     return x; 
}

static u16 mul_inv(u16 x) {    
       u16 t0, t1, q, y;         
       if (x < 2) 
          return x;    
       t0 = 0;    
       t1 = 65537 / x;    
       y  = 65537 % x;    
       while (y != 1) {       
             q = x / y;       
             x = x % y;       
             t0 = t0 + (t1 * q);       
             if (x == 1) 
                return t0;       
             q = y / x;       
             y = y % x;       
             t1 = t1 + (t0 * q);    
       }    
       return 65537-x; 
} 

static u16 addinv(u16 x) 
{    
     return 0-x; 
} 

//-------------------------------------------------------------------
// algoritmos do IDEA
//-------------------------------------------------------------------

// expande uma chave
static void expand_key(u16 *ukey, u16 *key ) 
{
       int i;    
       for (i=0; i<8; i++) 
           key[i]=ukey[i];    
       for (i=8; i<52; i++) {       
           if ((i & 7) < 6)          
              key[i]=(key[i-7] & 127) << 9 | key[i-6] >> 7;       
           else if ((i & 7) == 6)          
              key[i]=(key[i-7] & 127) << 9 | key[i-14] >> 7;       
           else          
              key[i]=(key[i-15] & 127) << 9 | key[i-14] >> 7;    
       } 
} 

// inverte a chave
void invert_key(u16 *in, u16 *out) 
{
       u16 t1, t2, t3, t4, round;    
       u16 *p = out + 52; 
       t1 = mul_inv(*in++);    
       t2 = addinv(*in++);    
       t3 = addinv(*in++);    
       t4 = mul_inv(*in++);    
       *--p = t4;    
       *--p = t3;    
       *--p = t2;    
       *--p = t1;    
       for (round = 1; round < 8; round++) {       
           t1 = *in++;       
           t2 = *in++;       
           *--p = t2;       
           *--p = t1;       
           t1 = mul_inv(*in++);       
           t2 = addinv(*in++);       
           t3 = addinv(*in++);       
           t4 = mul_inv(*in++);       
           *--p = t4;       
           *--p = t2; 
           *--p = t3;       
           *--p = t1;    
       }    
       t1 = *in++;    
       t2 = *in++;    
       *--p = t2;    
       *--p = t1;    
       t1 = mul_inv(*in++);    
       t2 = addinv(*in++);    
       t3 = addinv(*in++);    
       t4 = mul_inv(*in++);    
       *--p = t4;    
       *--p = t3;    
       *--p = t2;    
       *--p = t1; 
} 

// codigo auxiliar para cifrar/decifrar um bloco de bytes
int cipher( byte *outbuf, byte *inbuf, u16 *key )
{
    u16 x1, x2, x3,x4, s2, s3;
    u16 *in, *out;
    int r = IDEA_ROUNDS;

    in = (u16*)inbuf;
    x1 = *in++;
    x2 = *in++;
    x3 = *in++;
    x4 = *in;
    
    // se processador intel...
    #ifdef LITTLE_ENDIAN_HOST
	    x1 = (x1>>8) | (x1<<8);
	    x2 = (x2>>8) | (x2<<8);
	    x3 = (x3>>8) | (x3<<8);
	    x4 = (x4>>8) | (x4<<8);
  	#endif
  	
    do {
		MUL(x1, *key++);
		x2 += *key++;
		x3 += *key++;
		MUL(x4, *key++ );

		s3 = x3;
		x3 ^= x1;
		MUL(x3, *key++);
		s2 = x2;
		x2 ^=x4;
		x2 += x3;
		MUL(x2, *key++);
		x3 += x2;

		x1 ^= x2;
		x4 ^= x3;

		x2 ^= s3;
		x3 ^= s2;
    } while( --r );
    
    MUL(x1, *key++);
    x3 += *key++;
    x2 += *key++;
    MUL(x4, *key);

    out = (u16*)outbuf;
    
    // se processador intel...
  	#ifdef LITTLE_ENDIAN_HOST
	    *out++ = (x1>>8) | (x1<<8);
	    *out++ = (x3>>8) | (x3<<8);
	    *out++ = (x2>>8) | (x2<<8);
	    *out   = (x4>>8) | (x4<<8);
  	#else
	    *out++ = x1;
	    *out++ = x3;
	    *out++ = x2;
	    *out   = x4;
  	#endif
  	
  	return 0;
}

// define a chave atual
int do_setkey( IDEA_context *c, byte *key, int keylen )
{
    c->have_dk = 0;
    expand_key( (u16 *) key, (u16 *) c->ek );
    invert_key( c->ek, c->dk );
    
    return 0;
}

// cifra um bloco de bytes
void encrypt_block( IDEA_context *c, byte *outbuf, byte *inbuf )
{
    cipher( outbuf, inbuf, c->ek );
}

// decifra um bloco de bytes
void decrypt_block( IDEA_context *c, byte *outbuf, byte *inbuf )
{
    if( !c->have_dk ) { // flag de inverso da chave
       c->have_dk = 1;
       invert_key( c->ek, c->dk );
    } 
    
    cipher( outbuf, inbuf, c->dk );
}

//-------------------------------------------------------------------
// Cdigo principal
//-------------------------------------------------------------------

int idea_codificar(char *keystr, char *text, char *cript)
{
	int i, j, t;
	IDEA_context c;
	
    t = strlen(text);
    i = strlen(keystr);
    
	do_setkey(&c, (unsigned char*) keystr, i);
	
    for (j=0; j < t; j += 8) {
		encrypt_block(&c, (unsigned char*) cript, (unsigned char*) text);
	    text += 8;
	    cript += 8;
	}
	cript[0] = 0;
	
	return -1;
}

int idea_decodificar(char *keystr, char *cript, char *text, int size)
{
	int i, j;
	IDEA_context c;
	
    i = strlen(keystr);
    
	do_setkey(&c, (unsigned char*) keystr, i);
	
    for (j=0; j < size; j += 8) {
		decrypt_block(&c, (unsigned char*) text, (unsigned char*) cript);
	    text += 8;
	    cript += 8;
	}
	text[0] = 0;
	
	return -1;
}


