//-------------------------------------------------------------------
// Demonstrao de criptografia - 2007.1
// Projeto da cadeira de Topicos Especiais em Analise de Sistemas
// Equipe: Amaury Carvalho, 200610447
//		   Jousy Rodrigues, 200620260
//		   Flvio Mesquita, 200610222
//-------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "idea.h"
#include "des.h"
#include "rsa.h"
#include "md5.h"
#include "sha1.h"
#include "rc5.h"
#include "rl.h"

#define STRING_SIZE 255

char texto[STRING_SIZE] = "DEMONSTRACAO";
char cifra[STRING_SIZE] = "";
char chave[STRING_SIZE] = "123456789ABCDEF#";
char textodec[STRING_SIZE] = "";
int tamanhoTexto;
long tempo, vezes;

static void exibeCabec(char *titulo)
{
    system("cls");
  	printf("============================================\n");
  	printf("CIFRAGEM USANDO ALGORITMO %s\n", titulo);
  	printf("============================================\n");
}

static void exibeHexa(char *texto, int size)
{
  	unsigned char * caracteresCifra = (unsigned char*) texto;
  	int i;
  	for (i = 0; i < size; i++) { 
    	printf("%.2lX ", caracteresCifra[i]);
    }
}

bool novaConsulta() {
    char entrada[10];
	printf("\nNova consulta (S/N)? ");
    fflush(NULL); fflush(NULL);
	gets(entrada);
	return (entrada[0] == 'S' || entrada[0]=='s');
}

bool exibeResultado()
{
    double fator = fatorCorrelacao(texto, cifra, tamanhoTexto);
    if(fator < 0) fator *= -1;
    
  	printf("\nMensagem em texto claro........: %s", texto);
  	printf("\nMensagem em texto claro (ASCII): "); exibeHexa(texto, (tamanhoTexto>15)?15:tamanhoTexto);
	printf("\nMensagem cifrada (ASCII).......: "); exibeHexa(cifra, (tamanhoTexto>15)?15:tamanhoTexto);
    printf("\nMensagem decodificada..........: %s\n", textodec);
  
	printf("\nNumero de ciclos por segundo...: %d", vezes);
	printf("\nNumero de bytes por segundo....: %d", vezes * tamanhoTexto);
	printf("\nFator de correlacao do texto versus cifra: %.2f %%\n\n", fator*100);

    return novaConsulta();	
}

static void entraDados(int keySize)
{
    char entrada[STRING_SIZE];
    
    keySize /= 8;

    do {
       fflush(NULL); fflush(NULL);
       printf("Informe texto claro: ");
       gets(entrada);
       tamanhoTexto = strlen(entrada);
    } while(!tamanhoTexto);
    
    memset(texto, 0, STRING_SIZE);
    strcpy(texto, entrada);
    
    if(keySize) {
        do {
           fflush(NULL); fflush(NULL);
           printf("Informe a chave (%d caracteres): ", keySize);
           gets(entrada);
        } while(strlen(entrada) != keySize);
        
        strcpy(chave, entrada);
    }
}
       

int main( void ) 
{
    char entrada[10];
    int opcao = 0, volta;
    
    tamanhoTexto = strlen(texto);
    
    do {
        system("cls");
  	    printf("============================================\n");
        printf("DEMONSTRACAO DE CRIPTOGRAFIA\n");
  	    printf("============================================\n");
        printf("1) DES   (64/56)\n");
        printf("2) IDEA  (128/128)\n");
        printf("3) RSA   (64/128)\n");
        printf("4) MD5   (64/64)\n");
        printf("5) SHA-1 (64/64)\n");
        printf("6) RC5   (128/128)\n");
        printf("7) AES   (128/128)\n");
        printf("8) Todos\n");
        printf("0) Sair\n");
        
        printf("\n\nSelecione a opcao: ");
        fflush(NULL); fflush(NULL);
        gets(entrada);
        
        opcao = atoi(entrada);

        switch(opcao) {
            case 1 : {
                 do {
                     exibeCabec("DES");
                     entraDados(64);
                     vezes = 0;
      	             tempo = clock() + CLOCKS_PER_SEC;
                     while(clock() <= tempo) {
                        des_codificar(chave, texto, cifra);
          	            des_decodificar(chave, cifra, textodec, tamanhoTexto);
          	            vezes++;
                     }
       	         } while(exibeResultado());
            } break;
            case 2 : {
                 do {
                     exibeCabec("IDEA");
                     entraDados(128);
                     vezes = 0;
      	             tempo = clock() + CLOCKS_PER_SEC;
                     while(clock() <= tempo) {
                        idea_codificar(chave, texto, cifra);
                        idea_decodificar(chave, cifra, textodec, tamanhoTexto);
          	            vezes++;
                     }
       	         } while(exibeResultado());
            } break;
            case 3 : {
                 do {
                     char chavePublica[16], chavePrivada[16];
                     
                     exibeCabec("RSA");
                     entraDados(0);
                     
                     RSA_createKey(chavePublica, chavePrivada);
                     vezes = 0;
      	             tempo = clock() + CLOCKS_PER_SEC;
                     while(clock() <= tempo) {
                        rsa_codificar(chavePublica, texto, cifra);
                        rsa_decodificar(chavePrivada, cifra, textodec, tamanhoTexto);
          	            vezes++;
                     }
                     printf("\nATENCAO: esse algoritmo gera chaves automaticamente");
	                 printf("\nChave publica: "); exibeHexa(chavePublica, sizeof(chavePublica));
	                 printf("\nChave privada: "); exibeHexa(chavePrivada, sizeof(chavePrivada));
	                 printf("\n");
       	         } while(exibeResultado());
            } break;
            case 4 : {
                 char hash1[16], hash2[16];
                 if(!sha1_test()) printf("ERRO NO MD5!!!");
                 else
                     do {
                        exibeCabec("MD5");
                        entraDados(64);
                        md5_integridade(texto, hash1);
                        md5_autenticidade(chave, texto, hash2);
                        printf("\nHASH integridade..: "); exibeHexa(hash1, sizeof(hash1));
                        printf("\nHASH autenticidade: "); exibeHexa(hash2, sizeof(hash2));
                     } while(novaConsulta());
            } break;
            case 5 : {
                 char hash1[20], hash2[20];
                 if(!sha1_test()) printf("ERRO NO SHA1!!!");
                 else
                     do {
                        exibeCabec("SHA1");
                        entraDados(64);
                        sha1_integridade(texto, hash1);
                        sha1_autenticidade(chave, texto, hash2);
                        printf("\nHASH integridade..:"); exibeHexa(hash1, sizeof(hash1));
                        printf("\nHASH autenticidade:"); exibeHexa(hash2, sizeof(hash2));
                     } while(novaConsulta());
            } break;
            case 6 : {
                 do {
                     exibeCabec("RC5");
                     entraDados(128);
                     vezes = 0;
      	             tempo = clock() + CLOCKS_PER_SEC;
                     while(clock() <= tempo) {
                        rc5_codificar(chave, texto, cifra);
          	            rc5_decodificar(chave, cifra, textodec, tamanhoTexto);
          	            vezes++;
                     }
       	         } while(exibeResultado());
            } break;
            case 8 : {
                 exibeCabec("TODOS");
                 entraDados(64);
            } break;
        }
    } while(opcao > 0);
 
	return 0;
}

