/*------------------------------------------------------------
MD5
  O MD5 (Message-Digest algorithm 5)  um algoritmo de hash de 128 
  bits unidirecional desenvolvido pela RSA Data Security, Inc., 
  muito utilizado por softwares com protocolo ponto-a-ponto (P2P, 
  ou Peer-to-Peer, em ingls), verificao de integridade e logins.
  Foi desenvolvido em 1991 por Ronald Rivest para suceder ao MD4 
  que tinha alguns problemas de segurana. Por ser um algoritmo 
  unidirecional, uma hash md5 no pode ser transformada novamente no 
  texto que lhe deu origem. O mtodo de verificao , ento, feito 
  pela comparao das duas hash (uma da base de dados, e a outra da 
  tentativa de login). O MD5 tambem  usado pera verificar a 
  integridade de um ficheiro atravs, por exemplo, do programa 
  md5sum, que cria a hash de um ficheiro. Isto pode-se tornar muito 
  til para downloads de arquivos grandes, para programas P2P que 
  constroem arquivos atravs de pedaos e esto sujeitos  corrupo 
  dos mesmos. Como autenticao de login  utilizada em vrios sistemas 
  operacionais unix e em muitos sites com autentificao.

Funcionamento

Referencias:
  http://pt.wikipedia.org/wiki/MD5
  http://en.wikipedia.org/wiki/Md5
  http://en.wikipedia.org/wiki/Hmac
--------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "md5.h"

#define UCHAR unsigned char
#define UINT64 unsigned long long
#define UINT32 unsigned int

// adiciona com mdulo
static UINT32 addmod(UINT32 a, UINT32 b) {
   UINT64 c = a;
   c += b;
   return c & 0xFFFFFFFF; //% 0x100000000ll;
}
       
// faz a rotao  esquerda sem perda de bits
static UINT32 rol(UINT32 num, UINT32 cnt) {
  return (num << cnt) | (num >> (32 - cnt));
}

// copia um inteiro little endian para um buffer
void cpyInt(UCHAR *dest, UINT32 n) {
     dest[0] = n & 0xFF; n >>= 8;
     dest[1] = n & 0xFF; n >>= 8;
     dest[2] = n & 0xFF; n >>= 8;
     dest[3] = n & 0xFF; 
}

// l um inteiro little endian de um buffer
UINT32 readInt(UCHAR *src) {
     return *((UINT32 *) src);
}

// hash de integridade da mensagem
void hash_md5(UCHAR *message, int sizeOld, UCHAR *hash) {
     UCHAR filler[128], *buffer[2], *buf2;
     UINT32 w[16];
     UINT32 h0, h1, h2, h3;
     UINT32 a, b, c, d, e, f, g, temp;
     UINT32 sizeMsg = sizeOld, sizeFil, size[2];
     UINT32 i, j, m;
     UINT32 r[64] = {7, 12, 17, 22,  7, 12, 17, 22,  7, 12, 17, 22,  7, 12, 17, 22,
                     5,  9, 14, 20,  5,  9, 14, 20,  5,  9, 14, 20,  5,  9, 14, 20,
                     4, 11, 16, 23,  4, 11, 16, 23,  4, 11, 16, 23,  4, 11, 16, 23,
                     6, 10, 15, 21,  6, 10, 15, 21,  6, 10, 15, 21,  6, 10, 15, 21 };
     UINT32 k[64] = {
                  3614090360ul, 3905402710ul, 606105819ul,  3250441966ul,
                  4118548399ul, 1200080426ul, 2821735955ul, 4249261313ul,
                  1770035416ul, 2336552879ul, 4294925233ul, 2304563134ul,
                  1804603682ul, 4254626195ul, 2792965006ul, 1236535329ul,
                  4129170786ul, 3225465664ul, 643717713ul,  3921069994ul,
                  3593408605ul, 38016083ul,   3634488961ul, 3889429448ul,
                  568446438ul,  3275163606ul, 4107603335ul, 1163531501ul,
                  2850285829ul, 4243563512ul, 1735328473ul, 2368359562ul,
                  4294588738ul, 2272392833ul, 1839030562ul, 4259657740ul,
                  2763975236ul, 1272893353ul, 4139469664ul, 3200236656ul,
                  681279174ul,  3936430074ul, 3572445317ul, 76029189ul,
                  3654602809ul, 3873151461ul, 530742520ul,  3299628645ul,
                  4096336452ul, 1126891415ul, 2878612391ul, 4237533241ul,
                  1700485571ul, 2399980690ul, 4293915773ul, 2240044497ul,
                  1873313359ul, 4264355552ul, 2734768916ul, 1309151649ul,
                  4149444226ul, 3174756917ul, 718787259ul,  3951481745ul };
     
     // inicializa as variaveis:
     h0 = 0x67452301;
     h1 = 0xEFCDAB89;
     h2 = 0x98BADCFE;
     h3 = 0x10325476;

     // termina a mensagem original com o caracter 0x80
     // concatena 0s at que esta fique divisvel em blocos de 512 bits
     // sendo que os ultimos 64 bits devem conter o tamanho da mensagem (inteiro little endian)
     sizeFil = sizeMsg % 64;
     sizeMsg -= sizeFil;
     memset(filler, 0, sizeof(filler));
     memcpy(filler, message+sizeMsg, sizeFil);
     filler[sizeFil] = 0x80;  // terminador da mensagem
     if(sizeFil < 56) {
        sizeFil = sizeof(filler) / 2;
     } else {
        sizeFil = sizeof(filler);
     }
     cpyInt(filler+sizeFil-8, sizeOld*8); // grava o tamanho da mensagem (em bits) no final do ltimo bloco
     
     buffer[0] = message;
     buffer[1] = filler;
     size[0] = sizeMsg;
     size[1] = sizeFil;
     
     for(m = 0; m < 2; m++) {
           
         //processa a mensagem em blocos de 512 bits
         for(j = 0; j < size[m]; j+=64) {
    
               // copia os 512 bits do bloco corrente
               // dentro das 16 primeiras posies do
               // vetor de inteiros 32 bits (little endian)
               buf2 = &buffer[m][j];
               for(i=0; i < 16; i++) {
                  w[i] = readInt(buf2);
                  buf2 += 4;
               }
    
               // inicializa os valores hash do bloco
               a = h0;
               b = h1;
               c = h2;
               d = h3;
    
               // processa o vetor
               for(i = 0; i < 64; i++) {
                     if(i < 16) {
                          f = (b & c) | ((~b) & d);
                          g = i;
                     } else if(i < 32) {
                          f = (d & b) | ((~d) & c);
                          g = (5*i + 1) % 16;
                     } else if(i < 48) {
                          f = b ^ c ^ d;
                          g = (3*i + 5) % 16;
                     } else {
                          f = c ^ (b | (~d));
                          g = (7*i) % 16;
                     }
    
                     temp = d;
                     d = c;
                     c = b;
                     e = addmod(addmod(addmod(a, f), k[i]), w[g]);
                     b = addmod(b, rol(e, r[i]));
                     a = temp;
               }
            
               // adiciona o bloco ao hash global
               h0 = addmod(h0, a);
               h1 = addmod(h1, b);
               h2 = addmod(h2, c);
               h3 = addmod(h3, d);
         }
         
     }
     
     // produz o hash final de 160 bits (bigendian), 
     // hash = h0 concat h1 concat h2 concat h3 concat h4
     cpyInt(hash, h0); hash+=4;
     cpyInt(hash, h1); hash+=4;
     cpyInt(hash, h2); hash+=4;
     cpyInt(hash, h3); hash+=4;
        
}

// testa o algoritmo
int md5_test() {
    UCHAR hash[16];
    UCHAR test[] = "abc";
    UCHAR result[] = {0x90, 0x01, 0x50, 0x98, 0x3c, 0xd2, 0x4f, 0xb0, 0xd6, 0x96, 0x3f, 0x7d, 0x28, 0xe1, 0x7f, 0x72};
    hash_md5(test, sizeof(test), hash); 
    return strncmp((char *) result, (char *) hash, sizeof(hash));
}

// hash de autenticidade da mensagem
void hmac_md5(UCHAR *key, int sizeKey, UCHAR *message, int sizeMsg, UCHAR *hash) {
     UCHAR hkey[16];
     UCHAR opad[sizeof(hkey)], ipad[sizeof(hkey)];
     UCHAR *buf;
     int i;
     
     if(sizeMsg < 0) {
        memset(hash, 0, sizeof(hkey));
        return;
     }
     
     memset(opad, 0x5c, sizeof(opad));
     memset(ipad, 0x5c, sizeof(opad));
     
     if(sizeKey > sizeof(hkey)) {
        hash_md5(message, sizeMsg, hkey);
        sizeKey = sizeof(hkey);
     } else {
        memset(hkey, 0, sizeof(hkey));
        memcpy(hkey, key, sizeKey);
     }
       
     for(i=0; i < sizeKey; i++) {
        ipad[i] ^= hkey[i];
        opad[i] ^= hkey[i];
     }

     // concatena ipad com a mensagem
     buf = (UCHAR *) malloc(sizeof(ipad)+sizeMsg+100);
     if(buf) {
         memcpy(buf, ipad, sizeof(ipad));
         memcpy(buf+sizeof(ipad), message, sizeMsg);
         
         // calcula novo hash da chave
         hash_md5(buf, sizeof(ipad)+sizeMsg, hkey);
         
         // concatena opad com o novo hash
         memcpy(buf, opad, sizeof(opad));
         memcpy(buf+sizeof(opad), hkey, sizeof(hkey));
         
         // calcula o hash definitivo
         hash_md5(buf, sizeof(opad)+sizeof(hkey), hash);
         
         free(buf);
     }
}
 
//--------------------------------------------------------------
    
void md5_integridade(char *message, char *hash) {
     hash_md5((UCHAR *) message, strlen(message), (UCHAR *) hash);
}

void md5_autenticidade(char *key, char *message, char *hash) {
     hmac_md5((UCHAR *) key, strlen(key), (UCHAR *) message, strlen(message), (UCHAR *) hash);
}



